/** Diese Klasse reprsentiert Polynome dritter Ordnung der Form 
  * ax^3 + bx^2 + cx + d
  * Die Basisklasse ist die abstrakte Klasse Funktion, die abstrakte Methode
  * abbilden muss berschrieben werden.
  * 
  * @author Kirsten
  * @version 1.0
  */

public class Polynom extends Funktion
{
    //Die Koeffizienten a bis d des Polynoms
    private double koeffizientA;
    private double koeffizientB;
    private double koeffizientC;
    private double koeffizientD;

    /** Konstruktor, erzeugt leeres Polynom, da keine Daten bergeben wurden
     */
    public Polynom()
    {
	//ein Aufruf eines anderen Konstruktors muss als
	//erste Anweisung stehen
	this(0, 0, 0, 0);
    }
    
    /** Konstruktor, erzeugt Polynom 0. Ordnung, nur konstanter Wert
      * 
      * @param d Wert des konstanten Glieds
      */
    public Polynom(double d)
    {
	this(0, 0, 0, d);
    }

    /** Konstruktor, erzeugt Polynom 1. Ordnung, cx + d
      * 
      * @param c Wert des Koeffizienten c
      * @param d Wert des Koeffizienten d
      */
    public Polynom(double c, double d)
    {
	this(0, 0, c, d);
    }

    /** Konstruktor, erzeugt Polynom 1. Ordnung, bx^2 + cx + d
      * 
      * @param b Wert des Koeffizienten b
      * @param c Wert des Koeffizienten c
      * @param d Wert des Koeffizienten d
      */
    public Polynom(double b, double c, double d)
    {
        this(0, b, c, d);
    }

    /** Konstruktor, erzeugt Polynom 1. Ordnung, ax^3 + bx^2 + cx + d
      * 
      * @param a Wert des Koeffizienten a
      * @param b Wert des Koeffizienten b
      * @param c Wert des Koeffizienten c
      * @param d Wert des Koeffizienten d
      */
    public Polynom(double a, double b, double c, double d)
    {
	super();

	koeffizientA = a;
	koeffizientB = b;
	koeffizientC = c;
	koeffizientD = d;
    }

    /** Konstruktor, zum "Kopieren"
      * @param x Polynom, das kopiert werden soll
      */
    public Polynom(Polynom x)
    {
	this(x.koeffizientA, x.koeffizientB, x.koeffizientC, x.koeffizientD);
    }

    /** Methode zum Addieren zweier Polynome, d.h. aufaddieren der vier
     * Koeffizienten
     *
     * @param poly Polynom, dass zu einem anderen addiert werden soll
     * @return neues Polynom, in dem jeder Koeffizient der Addition der 
     *         beiden vorgegebenen Koeffizienten entspricht
     */
    public Polynom add(Polynom poly)
    {
	Polynom neu = new Polynom();

	neu.koeffizientA = this.koeffizientA + poly.koeffizientA;
	neu.koeffizientB = this.koeffizientB + poly.koeffizientB;
	neu.koeffizientC = this.koeffizientC + poly.koeffizientC;
	neu.koeffizientD = this.koeffizientD + poly.koeffizientD;

	return neu;
    }

    /** Methode zum Abbilden auf den Funktinswert, implementiert die abstrakte
     * Methode der Superklasse Funktion
     *
     * @param x Wert aus dem Wertebereich
     * @return Funktionswert zu x
     */
    public double abbilden(double x)
    {
	return koeffizientA * x * x * x + 
               koeffizientB * x * x +
	       koeffizientC * x + 
	       koeffizientD;
    }

    /** Methode, die die erste Ableitung des Polynoms liefert, also ein neues
     * Polynom, dessen Ordnung um eins niedriger ist
     *
     * @return erste Ableitung
     */
    public Polynom dx()
    {
	Polynom neu = new Polynom();

	neu.koeffizientA = 0;
	neu.koeffizientB = 3 * this.koeffizientA;
	neu.koeffizientC = 2 * this.koeffizientB;
	neu.koeffizientD = this.koeffizientC;

	return neu;
    }
    
    /** Methode, die das Integral zu einem Polynom liefert, wobei der 
     * konstante Faktor 0 ist
     *
     * @return Integral des Polynoms mit konstanten Faktor d = 0
     */
    public Polynom integral()
    {
	Polynom x = new Polynom();

	//da das Integral eine hhere Ordnung hat, darf das Ausgangs-
	//polynom maximal die Ordnung 2 haben
	if(this.koeffizientA == 0)
	{
	    x.koeffizientA = this.koeffizientB / 3;
	    x.koeffizientB = this.koeffizientC / 2;
	    x.koeffizientC = this.koeffizientD;
	    x.koeffizientD = 0;
	}
	
	return x;
    }

    /** Methode zur Ausgabe des Polynoms in einigermaen netter Form
     * berschreibt toString() von Object
     * 
     * @return Zeichenkette, die das Polynom "schn" darstellt
     */
    public String toString()
    {
	String ergebnis = "";

	if((koeffizientA == 0) && (koeffizientB == 0) && (koeffizientC == 0)
	   && (koeffizientD == 0))
	    ergebnis = "0";
	else
	{
	    if((koeffizientA != 0) && (koeffizientA != 1))
		ergebnis += koeffizientA + "*";
	    if(koeffizientA != 0)
	    {
		ergebnis += "x^3";
		if(!((koeffizientB == 0) && (koeffizientC == 0) && 
		     (koeffizientC == 0)))
		   ergebnis += " + ";
	    }

	    if((koeffizientB != 0) && (koeffizientB != 1))
		ergebnis += koeffizientB + "*";
	    if(koeffizientB != 0)
	    {
	        ergebnis += "x^2";
		if(!((koeffizientC == 0) && (koeffizientD == 0)))
		    ergebnis += " + ";
	    }

	    if((koeffizientC != 0) && (koeffizientC != 1))
		ergebnis += koeffizientC + "*";
	    if(koeffizientC != 0)
	    {
		ergebnis += "x";
		if(koeffizientD != 0)
		    ergebnis += " + ";
	    }

	    if(koeffizientD != 0)
		ergebnis += koeffizientD;
           
	}
	
	return ergebnis;
    }

    /** Methode zum Vergleichen von zwei Polynomen, berschreibt equals()
     * von Object()
     * Eine korrekte Implementierung ist: reflexiv (x.equals(x) -> true),
     * symmetrisch, transitiv, konsistent, x.equal(null) -> false.
     * 
     * @param x Object zum Vergleichen, Achtung: nicht Polynom als Typ 
     *          verwenden, sonst wird die Methode berschrieben, aber der
     *          eigentliche Vergleich funktioniert nicht!
     * @return true bei Gleichheit, false sonst
     */
    public boolean equals(Object x)
    {
	//der Obertyp ist abstrakt, ansonsten msste man auch
	//super().equals(x) prfen

	//auf jeden Fall muss der gleiche Typ vorliegen
	//instanceof ist bld bei Vererbung: drum
	if(this.getClass().equals(x.getClass()))
	{
	    if((this.koeffizientA == ((Polynom)x).getA()) &&
	       (this.koeffizientB == ((Polynom)x).getB()) &&
	       (this.koeffizientC == ((Polynom)x).getC()) &&
	       (this.koeffizientD == ((Polynom)x).getD()))
		return true;
	}
	
	return false;
    }

    /** Methode fr die Koeffizienten
     *
     * @return Koeffizient a
     */
    public double getA()
    {
	return koeffizientA;
    }

    /** Methode fr die Koeffizienten
     *
     * @return Koeffizient a
     */
    public double getB()
    {
	return koeffizientB;
    }

    /** Methode fr die Koeffizienten
     *
     * @return Koeffizient a
     */
    public double getC()
    {
	return koeffizientC;
    }

    /** Methode fr die Koeffizienten
     *
     * @return Koeffizient a
     */
    public double getD()
    {
	return koeffizientD;
    }
	
    /** Test, Test, Test, dies ist ein Test */
    public static void main(String args[])
    {
	Polynom p1 = new Polynom(5);
	Polynom p2 = new Polynom(2.5, 3);
	Polynom p3 = new Polynom(1.0, 0.0, 4.4);
	Polynom p4 = new Polynom(5.2, -2.1, 3.2, 0.0);

	System.out.println(p1);
	System.out.println(p2);
	System.out.println(p3);
	System.out.println(p4);

	System.out.println();

	System.out.println(p1.add(p2));
	System.out.println(p4.add(p2));
	System.out.println(p3.add(p2));
        System.out.println(p4.add(p3));

	System.out.println();

	System.out.println(p1.abbilden(5));
	System.out.println(p2.abbilden(5));
	System.out.println(p3.abbilden(5));
	System.out.println(p4.abbilden(5));

	System.out.println();

	System.out.println(p1.dx());
	System.out.println(p2.dx());
	System.out.println(p3.dx());
	System.out.println(p4.dx());

	System.out.println();
	
	System.out.println(p1.integral());
	System.out.println(p2.integral());
	System.out.println(p3.integral());
	System.out.println(p4.integral());

	System.out.println();
	
	Polynom p5 = new Polynom(p1.add(p2));
	Polynom p6 = new Polynom(p2.add(p1));

	System.out.println("Polynom 5 und 6 sind gleich: " + p5.equals(p6));
	System.out.println("Polynom 6 und 5 sind gleich: " + p6.equals(p5));
    }
}
